module Exercise05 where

import Data.List(partition)

import Data.Bifunctor

import Data.IORef
import Control.Monad

import System.Random

import Data.Tuple(swap)

import State

{-  Task 1: Monadic Quicksort
 
    Below you find a quick-sort implementation that works 
    with arbitrary monads that offer a getRandom- and an addCost-function.
    
    By designing suitable monads, you can select whether
    - costs should be ignored or should be calculated
    - the getRandom function is really backed by an rng, or whether it always returns 0

    By these choices you can obtain some variations of the quicksort algorithm.
-}

class Monad m => MonadCost m where
  addCost :: Integer -> m ()
  
class Monad m => MonadRandom m where
  getRandom :: Int -> m Int

qsortMonadic :: (MonadCost m, MonadRandom m, Ord a) => [a] -> m [a]
qsortMonadic [] = return []
qsortMonadic xs = do 
     pos <- getRandom $ n - 1
     let (xs1, x : xs2) = splitAt pos xs
     let (low, high) = partition (< x) (xs1 ++ xs2)
     addCost $ fromIntegral $ n - 1
     qs1 <- qsortMonadic low
     qs2 <- qsortMonadic high
     return $ qs1 ++ [x] ++ qs2
    where n = length xs
    
{-
The monad NoCostNoRandom should give you an idea, 
how to assemble everything. 
-}

newtype NoCostNoRandom a = NoCostNoRandom {runNCNR :: a}

  
returnNCNR = NoCostNoRandom
bindNCNR :: NoCostNoRandom a -> (a -> NoCostNoRandom b) -> NoCostNoRandom b
bindNCNR (NoCostNoRandom x) f = f x 

{- 
   the instances for Applicative, Functor and Monad are always the same, once
   you defined return and bind.
-}  

instance Functor NoCostNoRandom where
  fmap f x = x `bindNCNR` (\ y -> return $ f y)

instance Applicative NoCostNoRandom where
  pure = returnNCNR
  liftA2 f x1 g = x1 `bindNCNR` (\ x -> fmap (f x) g) 

instance Monad NoCostNoRandom where
  return = pure
  (>>=) = bindNCNR
  
  
-- trivial cost and random functions

instance MonadCost NoCostNoRandom where
  addCost _ = return ()
  
instance MonadRandom NoCostNoRandom where
  getRandom _ = return 0

-- obtain original qsort behavior
qsort :: Ord a => [a] -> [a]
qsort xs = runNCNR $ qsortMonadic xs

{-
   Task 1.A (2 points): 
   design a monad to obtain qsortCount from the lecture, 
   i.e., no rng, costs are counted
   
   Your defined should essentially be some wrapper around
   qsortMonadic, i.e., you may invoke qsortMonadic, but
   you are not allowed to modify qsortMonadic.
-}
newtype CostNoRandom a = Undefined a

returnCNR = undefined
bindCNR :: CostNoRandom a -> (a -> CostNoRandom b) -> CostNoRandom b
bindCNR x f = undefined

instance Functor CostNoRandom where
  fmap f x = x `bindCNR` (\ y -> return $ f y)
  
instance Applicative CostNoRandom where
  pure = returnCNR
  liftA2 f x1 g = x1 `bindCNR` (\ x -> fmap (f x) g) 

instance Monad CostNoRandom where
  return = pure
  (>>=) = bindCNR
  
instance MonadCost CostNoRandom where
  addCost n = undefined
  
instance MonadRandom CostNoRandom where
  getRandom _ = undefined

qsortCount :: Ord a => [a] -> (Integer, [a])
qsortCount xs = error "wrapper around (qsortMonadic xs)"


{-
   Task 1.B (3 points)
   design a monad to obtain a counting randomized quicksort function
   from qsortMonadic.
   The resulting function should be similar in behaviour to the
   one of qsortState, qsortIO, or qsortIOSeed from the lecture (your choice).
-}

data CRState = CRState { cost :: Integer, rng :: StdGen }

newtype CostRandom a = Undefined' a
  
returnCR = undefined
bindCR :: CostRandom a -> (a -> CostRandom b) -> CostRandom b
bindCR x f = undefined
  
instance Functor CostRandom where
  fmap f x = x `bindCR` (\ y -> return $ f y)
  
instance Applicative CostRandom where
  pure = returnCR
  liftA2 f x1 g = x1 `bindCR` (\ x -> fmap (f x) g) 

instance Monad CostRandom where
  return = pure
  (>>=) = bindCR
  
instance MonadCost CostRandom where
  addCost n = undefined

instance MonadRandom CostRandom where
  getRandom n = undefined

-- one of:
-- qsortCountRng :: Ord a => Int -> [a] -> (Integer, [a])
-- qsortCountRng :: Ord a => Int -> [a] -> IO (Integer, [a])
-- qsortCountRng :: Ord a => [a] -> IO (Integer, [a])
qsortCountRng = undefined 

{-
   Task 1.C (2 points)
   qsortMonadic has bad performance, if the list has lots of duplicates.

   Refactor qsortMonadic in a way that duplicates do not lead to inefficiency.

   Remark: Observe that all previous implementations profit from this refactoring.  
   
   You can test the efficiency using testListDups as input list.
-}


testListDups = concat $ replicate 6000 [1 .. 5 :: Int]

testList :: [Int]
testList = [1..12345]

testQS = putStrLn $ "sorted list of length: " ++ show (length (qsort testList))
testQSCount = putStrLn $ "comparisons: " ++ show (fst (qsortCount testList))

-- invocation depends on type of qsortCountRng;
-- adjustments might be required
--
-- testQSCountRng :: [Int] -> IO ()
-- testQSCountRng seeds = flip mapM_ seeds 
--   (\ seed -> return (qsortCountRng seed testList) >>= \ r -> print (fst r))
    
------------


{-
    Task 2 (3 points): Labeling terms
    
    We consider the same datatype from the exercise on the embedding relation.

    Define a function that assigns labels to a term.
    - the function should be written using a monadic style, using a suitable state-monad
    - every subterm should get its own number, where duplicates get separate numbers each
    - example: f(x,g(y),x) might be labeled as f0(x1,g2(y3),x4).
-}
data Term f v = Var v | Fun f [Term f v] deriving (Eq, Ord, Show)

type LTerm f v = Term (Int,f) (Int,v)

labelTerm :: Term f v -> LTerm f v
labelTerm t = undefined 


testTerm = Fun "f" [Var "x", Fun "g" [Var "y"], Var "x"]